// Licensed to the .NET Foundation under one or more agreements.
// The .NET Foundation licenses this file to you under the MIT license.

#ifdef FEATURE_SVR_GC
#define SERVER_GC 1
#include "gc.cpp"
#endif // FEATURE_SVR_GC
